#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <jni.h>
#include "vue_NativeVUE.h"

// NativeVUE class lookup data
static struct {
    jclass   clazz;   // Handle to NativeVUE class
    jfieldID pointer; // ID of NativeVUE.pointer field
} NativeVUE;

// Core context state type
typedef struct {
    uint8_t *rom;
    uint32_t rom_size;
} CORE;

// Retrieve the handle to a NativeVUE's core context
static CORE* GetCore(JNIEnv *env, jobject vue) {
    jbyteArray pointer =
        (*env)->GetObjectField(env, vue, NativeVUE.pointer);
    jbyte     *elems   = (*env)->GetByteArrayElements(env, pointer, NULL);
    CORE      *core    = *(CORE **)elems;
    (*env)->ReleaseByteArrayElements(env, pointer, elems, 0);
    return core;
}

// Native constructor
JNIEXPORT void JNICALL Java_vue_NativeVUE_construct
  (JNIEnv *env, jobject vue) {

    // Initialize the class data
    if (NativeVUE.clazz == NULL) {
        NativeVUE.clazz   = (*env)->GetObjectClass(env, vue);
        NativeVUE.pointer =
            (*env)->GetFieldID(env, NativeVUE.clazz, "pointer", "[B");
    }

    // Produce and initialize s new core context
    CORE *core = calloc(sizeof (CORE), 1);
    core->rom  = calloc(1024, 1);

    // Encode the context handle into a byte array
    jbyteArray pointer = (*env)->NewByteArray(env, sizeof (void *));
    jbyte     *elems   = (*env)->GetByteArrayElements(env, pointer, NULL);
    *(CORE **)elems = core;
    (*env)->ReleaseByteArrayElements(env, pointer, elems, 0);
    (*env)->SetObjectField(env, vue, NativeVUE.pointer, pointer);
}

// Release any used resources
JNIEXPORT void JNICALL Java_vue_NativeVUE_dispose
  (JNIEnv *env, jobject vue) {
    CORE *core = GetCore(env, vue);
    free(core->rom);
    free(core);
}

// Retrieve a copy of the ROM data
JNIEXPORT jbyteArray JNICALL Java_vue_NativeVUE_getROM
  (JNIEnv *env, jobject vue) {
    CORE *core = GetCore(env, vue);
    jbyteArray ret   = (*env)->NewByteArray(env, (jint) core->rom_size);
    jbyte     *elems = (*env)->GetByteArrayElements(env, ret, NULL);
    memcpy(elems, core->rom, core->rom_size);
    (*env)->ReleaseByteArrayElements(env, ret, elems, 0);
    return ret;
}

// Determine whether the context is native-backed
JNIEXPORT jboolean JNICALL Java_vue_NativeVUE_isNative
  (JNIEnv *env, jobject vue) {
    return JNI_TRUE;
}

// Provide new ROM data
JNIEXPORT jboolean JNICALL Java_vue_NativeVUE_setROM
  (JNIEnv *env, jobject vue, jbyteArray data, jint offset, jint length) {

    // Error checking
    if (data == NULL || offset < 0 || length < 1024 ||
        offset + length > (*env)->GetArrayLength(env, data) ||
        (length & length - 1) != 0)
        return JNI_FALSE;

    // Accept the new ROM data
    CORE *core = GetCore(env, vue);
    free(core->rom);
    core->rom = calloc(length, 1);
    jbyte *elems = (*env)->GetByteArrayElements(env, data, NULL);
    memcpy(core->rom, &elems[offset], length);
    (*env)->ReleaseByteArrayElements(env, data, elems, 0);
    core->rom_size = (uint32_t) length;
    return JNI_TRUE;
}
